# @(#) $Id: ssh-killer.bro 85 2009-02-14 20:41:22Z leres $ (LBL)
#
# Copyright (c) 2008, 2009
#      The Regents of the University of California.  All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that: (1) source code distributions
# retain the above copyright notice and this paragraph in its entirety, (2)
# distributions including binary code include the above copyright notice and
# this paragraph in its entirety in the documentation or other materials
# provided with the distribution, and (3) all advertising materials mentioning
# features or use of this software display the following acknowledgement:
# ``This product includes software developed by the University of California,
# Lawrence Berkeley Laboratory and its contributors.'' Neither the name of
# the University nor the names of its contributors may be used to endorse
# or promote products derived from this software without specific prior
# written permission.
# THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR IMPLIED
# WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
#

@load listen-clear
@load remote
@load notice
@load acld
@load email

# Broccoli linkage
redef Remote::destinations += {
	["syslog2bro"] = [
		$host = 127.0.0.1,
		$class = "syslog2bro",
		$events = /syslog2bro::(ssh_fail_login|ping)/,
		$connect=F,
		$ssl=F
	],
};

# Bind to to localhost
redef listen_if_clear = 127.0.0.1;

# Define new notices
redef enum Notice += {
	SshDebug,
	SshFailedLogin,
	SshFailedLoginBlocked,
	SshFailedLoginReBlockedAtInit,
	SshFailedLoginUnBlocked,
	SshFailedLoginBlockedAtExit,
	SshFailedLoginWhitelisted,
};

module ssh_killer;

export {
	## Globals
	global syslog2bro::ssh_fail_login: event(ts:time, src:addr,
	    dst:addr, user:string, method:string);

	global syslog2bro::ping: event(ts:time);
	global syslog2bro::pong: event(ts:time);

	## Tunables
	# How many failures before we block an attacker?
	const invalid_failures: count = 4 &redef;

	# How long do we wait before unblocking attackers?
	const blocked_timeout: interval = 2 days &redef;

	# How long do we keep track of attackers?
	# (Usually greater than blocked_timeout so we can immediately reblock)
	const invalid_timeout: interval = 4 day &redef;

	# Whitelist certain subnets
	const whitelist: set[subnet] = {
		127.0.0.0/8,
	} &redef;
}

# Forwards
global block: function(h: addr);
global unblock: function(h: addr);
global expire_blocked: function(t: set[addr], idx: addr): interval;

## Globals
global report: function(h: addr, block: bool);

# Re-block logic
global doing_init = T;

# Keep track of failures per address
global invalid: table[addr] of count
    &create_expire = invalid_timeout &default = 0 &persistent;

# Keep track of blocked hosts
global blocked: set[addr]
    &create_expire = blocked_timeout &expire_func = expire_blocked &persistent;

# Handler for ping
event syslog2bro::ping(ts:time)
{
	NOTICE([$note=SshDebug, $msg=fmt("ssh-killer: ping %s", ts) ]);
	event syslog2bro::pong(ts);
}

# Handler for failed ssh logins
event
syslog2bro::ssh_fail_login(ts:time, src:addr, dst:addr, user:string,
    method:string)
{
	NOTICE([$note=SshDebug,
	    $msg=fmt("ssh-killer: ssh_fail_login %s %s -> %s %s (%s)",
	    ts, src, dst, user, method) ]);

	# Check whitelist
	if (src in whitelist) {
		NOTICE([$note=SshFailedLoginWhitelisted, $src=src, $user=user,
		    $msg=fmt("whitelisted \"%s@%s\" (ignoring)", user, src) ]);
		return;
	}

	# Not interested in outbound failures
	if (is_local_addr(src))
		return;

	++invalid[src];

	NOTICE([$note=SshFailedLogin, $src=src, $user=user,
	    $msg=fmt("login failure \"%s@%s\"", user, src) ]);

	if (invalid[src] < invalid_failures)
		return;

	delete invalid[src];
	block(src);
}

# Block a host
function
block(h: addr)
{
	# No need to reblock unless starting up
	if (!doing_init && h in blocked)
		return;

	local why = SshFailedLoginBlocked;
	if (doing_init)
		why = SshFailedLoginReBlockedAtInit;
	NOTICE([$note=why, $src=h,
	    $msg=fmt("%sblocking %s", doing_init ? "re-" : "", h) ]);

	acld::acld_block(h);
	add blocked[h];
	if (!doing_init)
		report(h, T);
}

# Unblock a host
function
unblock(h: addr)
{
	NOTICE([$note=SshFailedLoginUnBlocked, $src=h,
	    $msg=fmt("unblock %s", h) ]);

	acld::acld_unblock(h);
	delete blocked[h];
	report(h, F);
}

# Expire blocked hosts
function
expire_blocked(t: set[addr], idx: addr): interval
{
	unblock(idx);
	return -1 secs;
}

# Hook to re-block hosts at startup
event
bro_init()
{
	for (h in blocked)
		block(h);
	doing_init = F;
}

#
# Hook to log blocked hosts at shutdown
#
# Note: bro_done() is called after persistent state is saved
#
event
bro_done()
{
	for (h in blocked)
		NOTICE([$note=SshFailedLoginBlockedAtExit, $src=h,
		    $msg=fmt("blocked at exit %s", h) ]);
}

# Report a newly blocked host
function
report(h: addr, block: bool)
{
	local now = current_time();
	local msg = fmt("%s %sblocked @ %D", h, block ? "" : "un", now);

	email::email(msg);
}
